@@ -26,13 +26,25 @@
 #include <string.h>
 #include <xenstore.h>
 #include <xenctrl.h>
+#include <poll.h>
 #include "libvchan.h"
 #include "libvchan_private.h"
 
+/* check if domain is still alive */
 int libvchan__check_domain_alive(xc_interface *xc_handle, int dom) {
     struct evtchn_status evst;
+    xc_dominfo_t dominfo;
     int ret;
-    /* check if domain still alive */
+
+    /* first try using domctl, more reliable but available in a privileged
+     * domain only */
+    ret = xc_domain_getinfo(xc_handle, dom, 1, &dominfo);
+    if (ret == 1)
+        return dominfo.domid == (uint32_t)dom && !dominfo.dying;
+    else if (ret == -1 && errno == ESRCH)
+        return 0;
+    /* otherwise fallback to xc_evtchn_status method */
+
     evst.dom = dom;
     /* xc_evtchn_status will return different error depending on
      * existence of "source" domain:
@@ -87,12 +99,11 @@ int libvchan_wait(libvchan_t *ctrl) {
          * right away).
          */
         while (ret == -2 && libxenvchan_is_open(ctrl->xenvchan) == 2) {
-            fd_set rd_set;
-            struct timeval tv = { 10, 0 };
             int vchan_fd = libxenvchan_fd_for_select(ctrl->xenvchan);
-            FD_ZERO(&rd_set);
-            FD_SET(vchan_fd, &rd_set);
-            switch (select(vchan_fd+1, &rd_set, NULL, NULL, &tv)) {
+            struct pollfd fds[] = {
+                { .fd = vchan_fd, .events = POLLIN, .revents = 0 },
+            };
+            switch (poll(fds, 1, 10000)) {
                 case 0:
                     if (!libvchan__check_domain_alive(ctrl->xc_handle, ctrl->remote_domain))
                         return -1;
@@ -104,7 +115,7 @@ int libvchan_wait(libvchan_t *ctrl) {
                 default:
                     if (errno == EINTR)
                         break;
-                    perror("select");
+                    perror("poll");
                     return -1;
             }
         }
@@ -170,6 +181,8 @@ void libvchan_close(libvchan_t *ctrl) {
         }
         free(ctrl->xs_path);
     }
+    if (ctrl->xc_handle)
+        xc_interface_close(ctrl->xc_handle);
     free(ctrl);
 }
 
